package eel6788;

import java.awt.geom.Rectangle2D;
import java.util.ArrayList;
import java.util.Random;

import javax.swing.JFrame;

import yaes.framework.simulation.IContext;
import yaes.framework.simulation.SimulationInput;
import yaes.framework.simulation.SimulationOutput;
import yaes.framework.world.World;
import yaes.framework.world.map.ArrangementHelper;
import yaes.framework.world.map.IObjectWithLocation;
import yaes.framework.world.map.Location;
import yaes.framework.world.map.PathGenerator;
import yaes.framework.world.map.PlannedPath;
import yaes.framework.world.sensornetwork.AbstractSensorAgent;
import yaes.framework.world.sensornetwork.ActuatorNode;
import yaes.framework.world.sensornetwork.SensorNetworkWorld;
import yaes.framework.world.sensornetwork.SensorNode;
import yaes.framework.world.sensornetwork.SinkNode;
import yaes.ui.propertysheet.DialogFrame;
import yaes.ui.propertysheet.PropertySheet;
import yaes.ui.visualization.VisualMap;
import yaes.ui.visualization.application.sensornetwork.MobileNodePainter;
import yaes.ui.visualization.application.sensornetwork.PainterConfiguration;
import yaes.ui.visualization.application.sensornetwork.SensorNetworkWorldPainter;
import yaes.ui.visualization.application.sensornetwork.TransmissionGroundMap;
import yaes.ui.visualization.application.sensornetwork.TransmissionGroundMapPainter;
import eel6788.agents.SimpleSensorAgent;
import eel6788.agents.directeddiffusion.DirectedDiffusionAgent;
import eel6788.agents.directeddiffusion.DirectedDiffusionSinkAgent;
import eel6788.gui.DirectedDiffusionNodePainter;
import eel6788.gui.DirectedDiffusionSinkNodePainter;

public class DirectedDiffusionContext implements IContext, Constants {

    private int sensorNodeCount;
    private int actuatorNodeCount;
    private TransmissionGroundMap map;
    private boolean visualize = true;
    private VisualMap visual;
    private PainterConfiguration painterConf;
    private int xMax;
    private int yMax;
    private final double speed = 1;
    private boolean inspectionEnabled = false;
    public static Random random = new Random();
    private SensorNetworkWorld sensorWorld;

    public DirectedDiffusionContext() {
    }

    /**
     * Distributes the sensor nodes according to the specification
     * 
     * @param sip
     */
    private void distributeSensorNodes(SimulationInput sip) {
        ArrayList<IObjectWithLocation> list2 = new ArrayList<IObjectWithLocation>(
                sensorWorld.getSensorNodes());
        switch (sip.getParameterEnum(SensorArrangement.class)) {
            case GRID : {
                ArrangementHelper.arrangeInAGrid(xMax / 4, yMax / 4,
                                                 3 * xMax / 4, 3 * yMax / 4,
                                                 list2);
                break;
            }
            case RANDOM : {
                ArrangementHelper
                        .arrangeRandomlyInARectangle(new Rectangle2D.Double(0,
                                0, xMax, yMax), list2, random);
                break;
            }
        }
    }

    /**
     * Creates a sensor agent for the given static node - based on data from the
     * sip. Some of them have special stuff.
     * 
     * @param sip
     * @param staticNode
     * @return
     */
    public AbstractSensorAgent createSensorNodeAgent(SimulationInput sip,
            SensorNode staticNode) {
        AbstractSensorAgent staticNodeAgent = null;
        switch (sip.getParameterEnum(SensorType.class)) {
            case SIMPLE : {
                final SimpleSensorAgent agent = new SimpleSensorAgent(
                        staticNode.getName(), sensorWorld);
                staticNodeAgent = agent;
                break;
            }
            case DIRECTED_DIFFUSION : {
                final DirectedDiffusionAgent agent = new DirectedDiffusionAgent(
                        staticNode.getName(), sensorWorld);
                staticNodeAgent = agent;
                break;
            }

        }
        staticNodeAgent.setTransmissionRange(sip
                .getParameterInteger(Constants.TRANSMISSION_RANGE));
        staticNodeAgent.setSensorRange(sip
                .getParameterInteger(Constants.SENSOR_RANGE));
        return staticNodeAgent;
    }

    /**
     * Initializes the environment
     * 
     */
    public void initialize(SimulationInput sip, SimulationOutput sop) {
        this.sensorWorld = new SensorNetworkWorld(sop);
        this.visualize = sip.getParameterEnum(VisualDisplay.class) == VisualDisplay.YES;
        this.sensorNodeCount = sip.getParameterInteger(Constants.SENSOR_NODES);
        this.actuatorNodeCount = sip
                .getParameterInteger(Constants.ACTUATOR_NODES);
        this.xMax = sip.getParameterInteger(Constants.XMAX);
        this.yMax = sip.getParameterInteger(Constants.YMAX);
        this.sensorWorld.setEndOfTheWorldTime((int) sip.getStopTime() - 1);
        map = new TransmissionGroundMap(xMax, yMax);
        // create the sink node
        final SinkNode sinkNode = new SinkNode();
        sinkNode.setName("Sink");
        final DirectedDiffusionSinkAgent sinkAgent = new DirectedDiffusionSinkAgent(
                "Sink", sensorWorld);
        sinkAgent.setTransmissionRange(sip
                .getParameterInteger(Constants.TRANSMISSION_RANGE));
        sinkNode.setAgent(sinkAgent);
        sinkAgent.setNode(sinkNode);
        sinkNode.setLocation(new Location(xMax / 2, yMax / 2));
        sensorWorld.getDirectory().addAgent(sinkAgent);
        sensorWorld.setSinkNode(sinkNode);
        // create and arrange the sensor nodes
        for (int i = 0; i < sensorNodeCount; i++) {
            final SensorNode staticNode = new SensorNode();
            staticNode.setName("StaticNode-" + i);
            final AbstractSensorAgent staticNodeAgent = createSensorNodeAgent(
                                                                              sip,
                                                                              staticNode);
            staticNode.setAgent(staticNodeAgent);
            staticNodeAgent.setNode(staticNode);
            sensorWorld.addSensorNode(staticNode);
            sensorWorld.getDirectory().addAgent(staticNodeAgent);
        }
        distributeSensorNodes(sip);
        // create an arrange the actuator nodes
        for (int i = 0; i < actuatorNodeCount; i++) {
            createMobileNode(i, sip);
        }
        // create the visual representations
        if (visualize) {
            createVisualRepresentation(sip);
        }
    }

    /**
     * Creates a mobile node
     */
    void createMobileNode(int i, SimulationInput sip) {
        PlannedPath path = null;
        final ActuatorNode mobileNode = new ActuatorNode(new Location(0, 0),
                "MobileNode-" + i, sensorWorld);
        ActuatorMovement value =  sip.getParameterEnum(ActuatorMovement.class);
        // the second actuator is always random, this is for testing purposes
        if (i==1) {
            value = ActuatorMovement.RANDOM_WAYPOINT;
        }
        switch (value) {
            case WINDING : {
                final int winds = 6;
                final int repetitions = 3;
                path = PathGenerator.createWindingPath(10, 10, xMax - 20, yMax
                        / winds, winds, repetitions);
                path.setSource(path
                        .getRandomLocation(DirectedDiffusionContext.random));
                break;
            }
            case RANDOM_WAYPOINT : {
                path = PathGenerator
                        .createRandomWaypointPathBySegments(
                                                            DirectedDiffusionContext.random,
                                                            new Rectangle2D.Double(
                                                                    0, 0, xMax,
                                                                    yMax), 50);
                break;
            }
            case ACCROSS : {
                ArrayList<Location> locations = new ArrayList<Location>();
                locations.add(new Location(0, 0));
                locations.add(new Location(xMax, yMax));
                path = PathGenerator.createPathFromNodes(locations);
            }
            case LSHAPED: {
                ArrayList<Location> locations = new ArrayList<Location>();
                locations.add(new Location(0, 0));
                locations.add(new Location(xMax / 2, yMax / 2));
                locations.add(new Location(xMax, 0));
                path = PathGenerator.createPathFromNodes(locations);                
            }
        }
        mobileNode.setMovement(path, speed);
        sensorWorld.addActuatorNode(mobileNode);
        sensorWorld.addObject(mobileNode);
    }

    /**
     * Creates a visual representation
     * 
     * @param sip
     */
    private void createVisualRepresentation(SimulationInput sip) {
        visual = new VisualMap(map.getWidth(), map.getHeight(), null,
                "Sensor network");
        painterConf = new PainterConfiguration();
        PropertySheet psheet = new PropertySheet(painterConf);
        DialogFrame dlgframe = new DialogFrame((JFrame) null, psheet);
        dlgframe.setVisible(true);
        // / visual.setScale(1.0);
        final TransmissionGroundMapPainter mapPainter = new TransmissionGroundMapPainter();
        visual.addObject(map, mapPainter);
        final DirectedDiffusionSinkNodePainter sinkPainter = new DirectedDiffusionSinkNodePainter(sensorWorld,
                painterConf);
        visual.addObject(sensorWorld.getSinkNode(), sinkPainter);
        final DirectedDiffusionNodePainter sPainter = new DirectedDiffusionNodePainter(sensorWorld,
                painterConf);
        for (final SensorNode node : sensorWorld.getSensorNodes()) {
            visual.addObject(node, sPainter);
        }
        final MobileNodePainter mPainter = new MobileNodePainter(sensorWorld,
                painterConf);
        for (final ActuatorNode node : sensorWorld.getActuatorNodes()) {
            visual.addObject(node, mPainter);
        }
        visual.addObject(sensorWorld,
                         new SensorNetworkWorldPainter(painterConf));
        // paint the path only if it is the winding one... otherwise it is
        // too complex
        /**
        if (path != null) {
            final PathPainter pPainter = new PathPainter();
            visual.addObject(path, pPainter);
            visual.setVisible(true);
        }
        */
    }

    public World getWorld() {
        return sensorWorld;
    }

    public void setVisualize(boolean visualize) {
        this.visualize = visualize;
    }

    public VisualMap getVisual() {
        return visual;
    }

    public TransmissionGroundMap getMap() {
        return map;
    }
}
